/**
 * 
 */
package com.platomix.userprofile.handler;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.platomix.userprofile.enums.DsEnum;
import com.platomix.userprofile.model.ShellCommond;
import com.platomix.userprofile.model.UserprofileTag;
import com.platomix.userprofile.service.DynamicExecuteService;
import com.platomix.userprofile.service.UserprofileTagService;
import com.platomix.userprofile.utils.ShellUtils;
import com.platomix.userprofile.utils.StrBuilder;
import com.platomix.userprofile.utils.TagKit;

import cn.hutool.core.date.StopWatch;
import cn.hutool.core.util.StrUtil;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;

/**
 * @author wookvn
 *
 *  根据临时表创建大宽表
 */
@Slf4j
@Service
public class FlatTagHandler {

	final static String CLUSTER_NAME = "cluster_3s_1r";
	final static String[] INVALID_FIELD_NAMES = new String[] { "USERID", "MONTH", "DEVICE_NUMBER", "cTime" };

	@Value("${dflow.flat-temp-table-name}")
	private String tempTableName;
	@Value("${dflow.flat-clu-table-name}")
	private String cluTableName;
	@Value("${dflow.flat-all-view-name}")
	private String flatAllViewName;
	
	@Value("${dbcluster.clickhouse}")
	private String[] clickhouseDbCluster;

	@Value("${dbcluster.clickhouse-etl}")
	private String clickhouseDbEtl;

	@Value("${spring.datasource.dynamic.datasource.clickhouse.username}")
	private String clickhouseUserName;
	@Value("${spring.datasource.dynamic.datasource.clickhouse.password}")
	private String clickhousePassWord;

	@Autowired
	private DynamicExecuteService dynamicExecuteService;
	
	@Autowired
	private UserprofileTagService userprofileTagService;
	
//	@SneakyThrows
//	public void test(String month) {
//		String tempColumnSql = "select name,`type` as tp from system.columns where `table`='" + TagKit.getTagTempTable(tempTableName,month) + "'";
//		List<ColumnParam> tempColumnList = dynamicExecuteService.list(DsEnum.clickhouse, tempColumnSql).stream().map(t -> new ColumnParam(t.getString("name"), t.getString("tp"))).collect(Collectors.toList());
//		//再次查询列变动后的字段名[入库以tempColumnList字段为主]
//		List<String> columnList = new ArrayList<>();
//		Map<String, String> columnTypeMap = new HashMap<>();
//		for(ColumnParam addColumn : tempColumnList) {
//			columnList.add("`" + addColumn.getName() + "`");
//			columnTypeMap.put("`" + addColumn.getName() + "`", addColumn.getType());
//		}
//		List<UserprofileTag> defaultTagList = userprofileTagService.list(Wrappers.<UserprofileTag>lambdaQuery().eq(UserprofileTag::getDefaultTagStatus, 1));
//		Map<String, UserprofileTag> defaultTagMap = _initDefaultTagMap(defaultTagList);
//		StrBuilder builder=new StrBuilder();
//		builder.append("select ")
//		.append(_dealSelectColumnSql(columnList, defaultTagMap, columnTypeMap))
//		.append(String.format(" from ads.%s where THEMONTH_V = toString(%s) and USERID_V>=toString(%s) and USERID_V<toString(%s)", TagKit.getTagTempTable(tempTableName,month),month, 13000, 13011));
////		builder.append("echo \"select ")
////		.append(_dealSelectColumnSql(columnList, defaultTagMap, columnTypeMap))
////		.append(String.format(" from ads.%s where THEMONTH_V = toString(%s) and USERID_V>=toString(%s) and USERID_V<toString(%s)\"", TagKit.getTagTempTable(tempTableName,month),month, 13000, 13011))
////		.append(String.format(" curl %s:8123?database=ads -u%s:%s -d @- | ",clickhouseDbEtl,clickhouseUserName,clickhousePassWord))
////		.append(String.format(" docker run -i --rm yandex/clickhouse-client --host %s --port 9007 --user %s --password %s --input_format_allow_errors_num=100",clickhouseDbCluster[0],clickhouseUserName,clickhousePassWord))
////		.append(String.format(" --query='INSERT INTO ads.%s( ",cluTableName))
////		.append(String.join(",", columnList))
////		.append(" ) FORMAT TabSeparated'");
//		FileUtils.writeStringToFile(new File("/Users/wookvn/Downloads/upload/todo.sh"), builder.toString(), "UTF-8");
//		System.out.println(builder.toString());
//	}
	
	/**
	 * 
	 * @param month
	 */
	public void tempDataToFlatView(String month) {
		//查询insert表和select表的字段，比对差异化
		String tempColumnSql = "select name,`type` as tp from system.columns where `table`='" + TagKit.getTagTempTable(tempTableName,month) + "'";
		List<ColumnParam> tempColumnList = dynamicExecuteService.list(DsEnum.clickhouse, tempColumnSql).stream().map(t -> new ColumnParam(t.getString("name"), t.getString("tp"))).collect(Collectors.toList());
		String clu1ColumnSql = "select name,`type` as tp from system.columns where `table`='" + cluTableName + "'";
		List<ColumnParam> clu1ColumnList = dynamicExecuteService.list(DsEnum.clickhouse91, clu1ColumnSql).stream().map(t -> new ColumnParam(t.getString("name"), t.getString("tp"))).collect(Collectors.toList());
		//获取需要删除和添加的字段
		List<ColumnParam> addColumnList = tempColumnList.parallelStream().filter(t -> !clu1ColumnList.stream().anyMatch(f -> f.getName().equals(t.getName()))).collect(Collectors.toList()),
				          delColumnList = clu1ColumnList.parallelStream().filter(t -> !tempColumnList.stream().anyMatch(f -> f.getName().equals(t.getName()))).collect(Collectors.toList());
		//校验并处理列变动
		_checkAndDealColumn(addColumnList, delColumnList);
		//删除月份分区数据
		String dropViewDataSql = String.format("alter table ads.%s on cluster %s drop partition '%s'", cluTableName, CLUSTER_NAME, month);
		log.info("\n" + dropViewDataSql);
		dynamicExecuteService.execute(DsEnum.clickhouse91, dropViewDataSql, "删除分区[month="+month+"]数据完成");
		try {
			TimeUnit.SECONDS.sleep(10);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		//再次查询列变动后的字段名[入库以tempColumnList字段为主]
		List<String> columnList = new ArrayList<>();
		Map<String, String> columnTypeMap = new HashMap<>();
		for(ColumnParam addColumn : tempColumnList) {
			columnList.add("`" + addColumn.getName() + "`");
			columnTypeMap.put("`" + addColumn.getName() + "`", addColumn.getType());
		}
		try {//生成shell文件再导入
			log.info("插入数据的字段列表:{}", String.join(",", columnList));
			shellToCluster(month,columnList,columnTypeMap);
		} catch (Exception e) {
			throw new RuntimeException(e.getMessage(), e);
		}
	}

	/**
	 * 手动生成shell文件，插入到不同服务器【只能在Linux下运行 需要clickhouse docker 环境】
	 * @param month 插入月份
	 * @param columnList 字段集合
	 * @throws Exception 
	 */
	private void shellToCluster(String month,List<String> columnList, Map<String, String> columnTypeMap) throws Exception{
		//按手机号 查找分段
		StopWatch stopWatch = new StopWatch("标签合并宽表时间");
		stopWatch.start("查询手机号段分布");
		Set<String> setps=new HashSet<>();
		StrBuilder sqlBuilder=new StrBuilder();
		log.info("开始查询手机号段分布区间");
		sqlBuilder.appendLn(String.format("with intDiv((select COUNT() from ads.%s),%d) as step", TagKit.getTagTempTable(tempTableName,month),clickhouseDbCluster.length))
				.appendLn("select USERID_V,(rowNumberInAllBlocks()+1) rownum from (")
				.appendLn(String.format("select USERID_V from ads.%s order by USERID_V ",TagKit.getTagTempTable(tempTableName,month))+")")
				.append("where rownum in (");
		IntStream.range(1, clickhouseDbCluster.length).forEachOrdered(idx -> {setps.add(String.format("step*%d",idx));});
		sqlBuilder.append(String.join(",", setps));
		sqlBuilder.append(")");
		//动态拼接SQL
		List<String> dataList = dynamicExecuteService.list(DsEnum.clickhouse, sqlBuilder.toString()).stream().map(item ->item.getString("USERID_V")).collect(Collectors.toList());
		stopWatch.stop();
		log.info("结束查询手机号段分布区间|length:{},cost:{}s", dataList.size(), stopWatch.getLastTaskTimeMillis()/1000.f);
		//初始化线程池，池子大小「clickhouseDbCluster#length」个
		ExecutorService threadPool = Executors.newFixedThreadPool(clickhouseDbCluster.length);
		//构造默认标签值数据map
		List<UserprofileTag> defaultTagList = userprofileTagService.list(Wrappers.<UserprofileTag>lambdaQuery().eq(UserprofileTag::getDefaultTagStatus, 1));
		Map<String, UserprofileTag> defaultTagMap = _initDefaultTagMap(defaultTagList);
		stopWatch.start("并发执行分片本地表数据插入");
		FileUtils.forceMkdir(new File(ShellUtils.SHELL_DIC));
		for(int i = 0; i < clickhouseDbCluster.length; i++) {
			String leftUserId=i==0?"0": dataList.get(i-1);
			String rightUserId=i==clickhouseDbCluster.length-1?"999": dataList.get(i);
			String userIdSql= leftUserId == null || rightUserId == null ? "" : String.format(" and USERID_V>=toString(%s) and USERID_V<toString(%s)",  leftUserId, rightUserId);

			StrBuilder builder=new StrBuilder();
			//命令大概如下：echo "select .." curl http://xxx -d @- | docker run .. --query='insert into xx() FORMAT TabSeparated'
			//clickhouse http请求方式使用参考：https://clickhouse.com/docs/zh/interfaces/http/
			//注意：前面是双引号，后面是单引号
			//"curl -d @-" 说明：-d 表示传递数据；@- 发送的内容来自标准输入（stdin）
			builder.append("echo \"select ")
					.append(_dealSelectColumnSql(columnList, defaultTagMap, columnTypeMap))
					.append(String.format(" from ads.%s where THEMONTH_V = toString(%s) %s\" | ",TagKit.getTagTempTable(tempTableName,month),month,userIdSql))
					.append(String.format(" curl %s:8123?database=ads -u%s:%s -d @- | ",clickhouseDbEtl,clickhouseUserName,clickhousePassWord))
					.append(String.format(" docker run -i --rm yandex/clickhouse-client --host %s --port 9007 --user %s --password %s --input_format_allow_errors_num=100",clickhouseDbCluster[i],clickhouseUserName,clickhousePassWord))
					.append(String.format(" --query='INSERT INTO ads.%s( ",cluTableName))
					.append(String.join(",", columnList))
					.append(" ) FORMAT TabSeparated'");
			ShellCommond shellCommand = new ShellCommond(clickhouseDbCluster[i],false,false,builder.toString());
			//添加任务到线程池中执行
			threadPool.execute(() -> {
				log.info("====开始执行[{}]数据写入====", shellCommand.getIp());
				long startTime = System.currentTimeMillis();
				try {
					ShellUtils.shellToCluster(shellCommand);
				} catch (Exception e) {
					log.error(e.getMessage());
				}
				log.info("====结束执行[{}]数据写入|cost:{}s====", shellCommand.getIp(), (System.currentTimeMillis()-startTime)/1000.f);
			});
		}
		//发送一个关闭信号「awaitTermination」这个方法接收
		threadPool.shutdown();
		//阻塞任务执行，最多执行1个小时
		boolean ret = threadPool.awaitTermination(1, TimeUnit.HOURS);
		if(!ret) { //如果没有正常结束，说明还有任务在执行，强制关闭
			threadPool.shutdownNow();
		}
		stopWatch.stop();
		log.info("合并宽表完成,总计耗时:{}s", stopWatch.getLastTaskTimeMillis()/1000.f);
	}
	
	/**
	 * 创建Flat表 [初始化执行一次]
	 */
	public void initCreateFlatTable(String month) {
		List<JSONObject> columnList = dynamicExecuteService.list(DsEnum.clickhouse, "select name,`type` from system.columns where `table`='" + TagKit.getTagTempTable(tempTableName,month) + "'");
		StrBuilder sqlBuilder = new StrBuilder();
		sqlBuilder.appendLn("create table ads." + cluTableName + " on cluster " + CLUSTER_NAME + " (")
				  .appendLn("`THEMONTH_V` String, ").appendLn("`USERID_V` String,");
		for (JSONObject column : columnList) {
			String name = column.getString("name"), type = column.getString("type");
			if (!Stream.of(INVALID_FIELD_NAMES).anyMatch(f -> name.contains(f))) {
				sqlBuilder.append("`").append(name).append("` ").append(type).appendLn(",");
			}
		}
		// 删除一个空格和一个逗号
		sqlBuilder.deleteLastChar().deleteLastChar();
		sqlBuilder.appendLn(") ENGINE = MergeTree PARTITION BY THEMONTH_V ORDER BY USERID_V");
		log.info("\n" + sqlBuilder.toString());
		dynamicExecuteService.execute(DsEnum.clickhouse91, sqlBuilder.toString(), "创建ads.Z_USER_TAG_FLAT_clu表完成");
	}

	/**
	 * 校验字段列是否有改变
	 */
	private void _checkAndDealColumn(List<ColumnParam> addColumnList, List<ColumnParam> delColumnList) {
		List<String> addColumns = new ArrayList<>();
		List<String> delColumns = new ArrayList<>();
		for(ColumnParam addColumn : addColumnList) {
			addColumns.add("add column `" + addColumn.getName() + "` " + addColumn.getType());
		}
		for(ColumnParam addColumn : delColumnList) {
			delColumns.add("drop column `"+addColumn.getName()+"`");
		}
		
		//是否需要增删字段； 是->先drop视图
		if(!addColumns.isEmpty() || !delColumns.isEmpty()) {
			//1.drop视图
			String dropViewSql = String.format("drop table IF EXISTS %s ON cluster %s", flatAllViewName, CLUSTER_NAME);
			log.info("\n" + dropViewSql);
			dynamicExecuteService.execute(DsEnum.clickhouse91, dropViewSql, "删除视图完成");
			log.info("删除视图完成，开始等待8s处理增删列...");
			try {
				TimeUnit.SECONDS.sleep(8);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			
			//2.处理增删列
			_dealColumn(addColumns);
			_dealColumn(delColumns);
			
			//3.创建视图
			//CREATE TABLE ads.Z_USER_TAG_FLAT_ALL on cluster cluster_3s_1r as ads.Z_USER_TAG_FLAT_clu1 ENGINE = Distributed('cluster_3s_1r', 'ads', 'Z_USER_TAG_FLAT_clu1', rand())
			String createViewSql = String.format("CREATE TABLE %s ON cluster %s AS ads.%s ENGINE = Distributed('%s', 'ads', '%s', rand())", 
					flatAllViewName, CLUSTER_NAME, cluTableName, CLUSTER_NAME, cluTableName);
			log.info("\n" + createViewSql);
			dynamicExecuteService.execute(DsEnum.clickhouse91, createViewSql, "创建视图完成");
			log.info("创建视图完成，开始等待8s进入下一操作...");
			try {
				TimeUnit.SECONDS.sleep(8);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * 处理字段列
	 * @param columns
	 */
	private void _dealColumn(List<String> columns) {
		if(!columns.isEmpty()) {
			String addColumnSql = String.format("alter table ads.%s ON cluster %s %s", cluTableName, CLUSTER_NAME, String.join(",", columns));
			log.info("\n" + addColumnSql);
			dynamicExecuteService.execute(DsEnum.clickhouse91, addColumnSql, "更新字段完成");
		}
	}
	
	/**
	 * 
	 * @param columnList
	 * @return
	 */
	private String _dealSelectColumnSql(List<String> columnList, Map<String, UserprofileTag> defaultTagMap, Map<String, String> columnTypeMap) {
		StrBuilder sqlBuilder = new StrBuilder();
		columnList.forEach(columnName -> _handleAllTags(sqlBuilder, columnName, defaultTagMap, columnTypeMap));
		sqlBuilder.deleteLastChar();
		return sqlBuilder.toString();
	}
	
	/**
	 * 处理未打标签设置默认标签值
	 * @param sqlBuilder
	 * @param columnName
	 * @param defaultTagMap
	 * 说明： 为什么会有反斜线。 主要是因为linux命令中：双引号，单引号，反引号的问题。
	 * 1、在双引号中，反引号中的内容是命令；
	 * 2、在单引号中，反引号中的内容是字符串。
	 */
	private void _handleAllTags(StrBuilder sqlBuilder, String columnName, Map<String, UserprofileTag> defaultTagMap, Map<String, String> columnTypeMap) {
		UserprofileTag tag = defaultTagMap.get(columnName);
		String _columnName = "\\" + StrUtil.subPre(columnName, -1) + "\\`";
		if(tag != null) {
			//判断是否数组
			boolean isArray = "array".equalsIgnoreCase(tag.getTagCType());
			if(isArray) {
				if(columnName.endsWith(".E`")) {
					sqlBuilder.append(String.format("if(empty(%s), ['%s'], %s) as %s,", _columnName, tag.getTagId(), _columnName, _columnName));
				}else if(columnName.endsWith(".C`")) {
					sqlBuilder.append(String.format("if(empty(%s), ['%s'], %s) as %s,", _columnName, tag.getTagName(), _columnName, _columnName));
				}else {
					boolean isNumber = !"Array(String)".equalsIgnoreCase(columnTypeMap.get(columnName));
					sqlBuilder.append(String.format("if(empty(%s), [%s], %s) as %s,", _columnName, isNumber ? "0" : "''", _columnName, _columnName));
				}
			}else {
				if(columnName.endsWith("_V`")) {
					sqlBuilder.append(_columnName + ",");
				}else if(columnName.endsWith("_C`")) {
					sqlBuilder.append(String.format("if(isNull(%s) or empty(%s), '%s', %s) as %s,", _columnName, _columnName, tag.getTagName(), _columnName, _columnName));
				}else {
					sqlBuilder.append(String.format("if(isNull(%s) or empty(%s), '%s', %s) as %s,", _columnName, _columnName, tag.getTagId(), _columnName, _columnName));
				}
			}
		}else {
			sqlBuilder.append(_columnName + ",");
		}
	}
	
	/**
	 * 处理标签
	 * @param sqlBuilder
	 * @param columnName
	 * @return
	 */
//	private boolean _handleComplain(StrBuilder sqlBuilder, String columnName) {
//		if(columnName.contains("`TAG_COMPLAIN`")) {
//			sqlBuilder.append("ifNull("+columnName+", '\"'\"'B112_22_001'\"'\"') as `TAG_COMPLAIN`,");
//			return true;
//		}else if(columnName.contains("`TAG_COMPLAIN_C`")) {
//			sqlBuilder.append("ifNull("+columnName+", '\"'\"'0次'\"'\"') as `TAG_COMPLAIN_C`,");
//			return true;
//		}
//		return false;
//	}
	
	/**
	 * 初始化默认标签值
	 * @param defaultTagList
	 * @return
	 */
	private Map<String, UserprofileTag> _initDefaultTagMap(List<UserprofileTag> defaultTagList){
		Map<String, UserprofileTag> defaultTagMap = new HashMap<>();
		for(UserprofileTag tag : defaultTagList) {
			String tagEname = "array".equalsIgnoreCase(tag.getTagCType()) ? tag.getEnglishName().toUpperCase() + ".E" : tag.getEnglishName().toUpperCase(),
				   tagCname = "array".equalsIgnoreCase(tag.getTagCType()) ? tag.getEnglishName().toUpperCase() + ".C" : tag.getEnglishName().toUpperCase() + "_C",
				   tagVname = "array".equalsIgnoreCase(tag.getTagCType()) ? tag.getEnglishName().toUpperCase() + ".V" : tag.getEnglishName().toUpperCase() + "_V";
			defaultTagMap.put("`" + tagEname + "`", tag);
			defaultTagMap.put("`" + tagCname + "`", tag);
			defaultTagMap.put("`" + tagVname + "`", tag);
		}
		return defaultTagMap;
	}
	
	@Data
	@AllArgsConstructor
	static class ColumnParam{
		private String name;
		private String type;
	}
	
}
