/**
 * 
 */
package com.platomix.userprofile.handler.core;

import java.time.LocalDate;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.platomix.userprofile.enums.DsEnum;
import com.platomix.userprofile.handler.AbstractTagHandler;
import com.platomix.userprofile.model.UserprofileTag;
import com.platomix.userprofile.service.UserprofileTagService;
import com.platomix.userprofile.utils.StrBuilder;

import lombok.extern.slf4j.Slf4j;

/**
 * @author wookvn
 *
 *	生成全量用户数据，包括信息化用户和订购关系用户的并集 【该表跟月份没有关系】
 *  可以理解为联通的全量用户表
 */
@Slf4j
@Service
public class UnicomLongTagHandler extends AbstractTagHandler {
	final static String UNICOM_TABLE_NAME = "ads.UNICOM_ALL_MONTH_LONG";
	final static String ZT_TABLE_NAME = "ads.ZT_USER_TAG_LONG";

	@Autowired
	private UserprofileTagService userprofileTagService;

	@Override
	public void invoke(String month, String mobileLeft, String mobileRight) {
		String lrMobileSql = mobileLeft == null || mobileRight == null ? "" : String.format(" and DEVICE_NUMBER>='%s' and DEVICE_NUMBER<'%s'", mobileLeft, mobileRight),
			   lrUseridSql = mobileLeft == null || mobileRight == null ? "" : String.format(" and USERID>='%s' and USERID<'%s'", mobileLeft, mobileRight);
		StrBuilder sqlBuilder = new StrBuilder();
		int year = LocalDate.now().getYear();
		sqlBuilder.appendLn(super.buildTableSql(UNICOM_TABLE_NAME, null, "DEVICE_NUMBER"));
		sqlBuilder.appendLn("select DEVICE_NUMBER,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x), groupArray(PROV_DESC)) PROV_DESC,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x), groupArray(SERVICE_TYPE)) SERVICE_TYPE,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x), groupArray(CUST_SEX)) CUST_SEX,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x), groupArray(CERT_AGE)) CERT_AGE,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x), groupArray(PRODUCT_CLASS_DESC)) PRODUCT_CLASS_DESC,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x), groupArray(MON_FEE)) MON_FEE");
		sqlBuilder.appendLn("from ");
		sqlBuilder.appendLn("(select DEVICE_NUMBER1 as DEVICE_NUMBER,if(empty(s3.PROV_DESC) or isNull(s3.PROV_DESC),b.PROVINCE,s3.PROV_DESC) as PROV_DESC,SERVICE_TYPE,")
				  .appendLn("CUST_SEX,CERT_AGE,s3.PRODUCT_CLASS_DESC as PRODUCT_CLASS_DESC,c.MON_FEE as MON_FEE ");
		sqlBuilder.appendLn("from ( ");
		sqlBuilder.appendLn("select s1.*, if(empty(s1.DEVICE_NUMBER), s2.USERID, s1.DEVICE_NUMBER) DEVICE_NUMBER1 from");
		sqlBuilder.appendLn("(select MONTH_ID, DEVICE_NUMBER ,PROV_DESC,SERVICE_TYPE ,CUST_SEX,CERT_AGE,PRODUCT_CLASS_DESC from ads.XXH_ALL_MONTH ");
		sqlBuilder.appendLn(String.format("	where  MONTH_ID >='202101' and  DEVICE_NUMBER not in (select DEVICE_NUMBER from ads.XXH_KUAISHOU_MONTH where 1=1 %s) %s ",lrMobileSql,lrMobileSql));
		sqlBuilder.appendLn("  union all ");
		sqlBuilder.appendLn(" select MONTH_ID,DEVICE_NUMBER, PROV_DESC, NULL as SERVICE_TYPE,if(CUST_SEX==1,'男','女') as CUST_SEX, ");
		sqlBuilder.appendLn("  multiIf(("+year+"-CERT_AGE)>=2010,'10后',("+year+"-CERT_AGE)>=2000,'00后',("+year+"-CERT_AGE)>=1990,'90后',("+year+"-CERT_AGE)>=1980,'80后', ");
		sqlBuilder.appendLn("  ("+year+"-CERT_AGE)>=1970,'70后',("+year+"-CERT_AGE)<1970,'70前','未知') as CERT_AGE, ");
		sqlBuilder.appendLn("  PRODUCT_CLASS_DESC ");
		sqlBuilder.appendLn("  from ads.XXH_KUAISHOU_MONTH where 1=1 " + lrMobileSql);
		sqlBuilder.appendLn(String.format("  ) s1 full outer join (select distinct USERID from ETL_TD_ORDER_RELATIONS_YEAR where THEMONTH > '202101' %s) s2 on s1.DEVICE_NUMBER = s2.USERID", lrUseridSql));
		sqlBuilder.appendLn(")s3");
		sqlBuilder.appendLn("  left join  ads.ETL_VW_TB_CALLER_AREACODE b on substring(s3.DEVICE_NUMBER1, 1, 7) = b.CALLERHEAD ");
		//--aokunsang 20220710 套餐价格
		sqlBuilder.appendLn("  left join ads.DIM_PACKAGE_PRICE c on s3.PRODUCT_CLASS_DESC = c.PRODUCT_CLASS_DESC ");
		//sqlBuilder.appendLn("	order by MONTH_ID desc ");
		sqlBuilder.appendLn(") group by DEVICE_NUMBER ");
//		log.info("\n" + sqlBuilder.toString());
		dynamicExecuteService.execute(DsEnum.clickhouse, sqlBuilder.toString(), String.format("[%s]Long标签数据缓存表生成{%s-%s}", UNICOM_TABLE_NAME, mobileLeft, mobileRight));
	}
	
	@Override
	public boolean executeInBatchs() { 
		return true; 
	}

	/**
	 * before invoke 执行
	 * @return
	 */
	@Override
	public void beforeInvoke(String month) {
		dynamicExecuteService.execute(DsEnum.clickhouse, String.format("drop table if exists %s",UNICOM_TABLE_NAME), "删除long标签缓存表数据");
		dynamicExecuteService.execute(DsEnum.clickhouse, String.format("drop table if exists %s",ZT_TABLE_NAME), "删除long标签中间表数据");
		try {
			TimeUnit.SECONDS.sleep(10);
		} catch (InterruptedException e) {
			log.error(e.getMessage(), e);
		}
	}

	/**
	 * after invoke 执行
	 * @return
	 */
	@Override
	public void afterInvoke(String month){
		try {
			TimeUnit.SECONDS.sleep(10);
			log.info("开始从缓存表生成LONG标签中间表数据|tableName={}", ZT_TABLE_NAME);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

		StrBuilder sqlBuilder = new StrBuilder();
		sqlBuilder.appendLn(super.buildTableSql(ZT_TABLE_NAME, null, "DEVICE_NUMBER_XXH_L"));
		sqlBuilder.appendLn("select DEVICE_NUMBER as DEVICE_NUMBER_XXH_L,");
		//性别
		String sexSql = userprofileTagService.genSelectSql("性别", "CUST_SEX", null, null);
		sqlBuilder.append(sexSql).appendLn(",");

		//省份
		String provinceSql = userprofileTagService.genSelectSql("省份", "PROV_DESC", null, null);
		sqlBuilder.append(provinceSql).appendLn(",");

		//网络类型
		String netSql = userprofileTagService.genSelectSql("网络类型", "SERVICE_TYPE", null, null);
		sqlBuilder.append(netSql).appendLn(",");

		//年龄阶段
		String ageSql = userprofileTagService.genSelectSql("年龄阶段", "CERT_AGE", null, null);
		sqlBuilder.appendLn(ageSql).appendLn(",");
		
		//套餐价格 20220710 从XxhTagHandler迁移过来
		List<UserprofileTag> tagList = userprofileTagService.list(Wrappers.lambdaQuery(UserprofileTag.class).eq(UserprofileTag::getEnglishName, "tag_package_pic"));
		Map<String, List<UserprofileTag>> tagListMap = tagList.stream().collect(Collectors.groupingBy(t -> t.getEnglishName().toLowerCase()));
		String sql = genBucketSql("tag_package_pic", "toFloat32OrNull(MON_FEE)", "toFloat32OrNull(MON_FEE)", tagListMap.get("tag_package_pic"));
		sqlBuilder.appendLn(sql);

		sqlBuilder.appendLn(String.format("from %s ",UNICOM_TABLE_NAME));
//		log.info("\n" + sqlBuilder.toString());
		dynamicExecuteService.execute(DsEnum.clickhouse, sqlBuilder.toString(), "["+ZT_TABLE_NAME+"]Long标签中间表生成");
	}
	
	@Override
	public String tableName(String month) {
		return ZT_TABLE_NAME;
	}
}
