/**
 * 
 */
package com.platomix.userprofile.handler.core;

import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.platomix.userprofile.enums.DsEnum;
import com.platomix.userprofile.handler.AbstractTagHandler;
import com.platomix.userprofile.model.UserprofileTag;
import com.platomix.userprofile.service.UserprofileTagService;
import com.platomix.userprofile.utils.StrBuilder;

/**
 * @author wookvn
 *
 *	是否关注公众号/是否关注小程序
 */
//@Slf4j
@Component
public class WeixinTagHandler extends AbstractTagHandler {

	final static String ZT_TABLE_NAME = "ads.ZT_USER_TAG_WEIXIN";
	
	@Autowired
	private UserprofileTagService userprofileTagService;
	
	@Override
	public void beforeInvoke(String month) {
		boolean isExists = dynamicExecuteService.checkTableExistsOfClickhouse(ZT_TABLE_NAME);
		if(isExists) { //如果表存在，清空数据重新跑【该表没有分区】
			String dropViewDataSql = String.format("truncate table %s", ZT_TABLE_NAME);
			dynamicExecuteService.execute(DsEnum.clickhouse, dropViewDataSql, "清除[table="+ZT_TABLE_NAME+"]数据完成");
			try {
				TimeUnit.SECONDS.sleep(10);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	@Override
	public void invoke(String month, String mobileLeft, String mobileRight) {
		// 查询所有标签
		List<UserprofileTag> tagList = userprofileTagService.list();
		Map<String, List<UserprofileTag>> tagListMap = tagList.stream().collect(Collectors.groupingBy(t -> t.getEnglishName().toLowerCase()));
		StrBuilder sqlBuilder = new StrBuilder();
		sqlBuilder.appendLn(super.buildTableSql(ZT_TABLE_NAME, "USERID_WX"));
		sqlBuilder.append("select ").append("if(mobile = '', bmobile, mobile) as USERID_WX,");
		//关注公众号
		String subscribeSql = genBucketSql("tag_wxpublic_follow", "subscribe", "subscribe", tagListMap.get("tag_wxpublic_follow"));
		sqlBuilder.append(subscribeSql).appendLn(",");
		//绑定小程序
		String bindSql = genBucketSql("tag_wxapp_bind", "binding_time", "binding_time", tagListMap.get("tag_wxapp_bind"));
		sqlBuilder.appendLn(bindSql);
		
		sqlBuilder.appendLn("from (")
				  .appendLn("select")
				  .appendLn("a.*, b.mobile as bmobile, b.subscribe, b.subscribe_time, b.unsubscribe_time") 
				  .appendLn("from ads.ETL_SYNC_FLOW_USER_MOBILE_OPENID a")
				  .appendLn("full outer join ads.ETL_SYNC_TB_WX_USER_INFO b on a.mobile = b.mobile")
			      .appendLn(")");
//		log.info("\n" + sqlBuilder.toString());
		dynamicExecuteService.execute(DsEnum.clickhouse, sqlBuilder.toString(), String.format("[%s]微信关注公众号和绑定小程序相关标签生成", ZT_TABLE_NAME));
	}
	
	@Override
	public String tableName(String month) {
		return ZT_TABLE_NAME;
	}

}
