/**
 * 
 */
package com.platomix.userprofile.handler.support;

import org.springframework.stereotype.Component;

import com.platomix.userprofile.enums.DsEnum;
import com.platomix.userprofile.handler.AbstractTagHandler;
import com.platomix.userprofile.utils.StrBuilder;

/**
 * @author wookvn
 *
 *	跑数据使用 - 用户部分标签的最新数据
 */
//@Slf4j
@Component
public class UserLatestTagManager extends AbstractTagHandler {

	final static String ZT_TABLE_NAME = "ads.ZT_USER_PARTTAG";

	final static long _STEP = 20000000L;
	
	@Override
	public String tableName(String month) {
		return ZT_TABLE_NAME;
	}

	@Override
	public void invoke(String month, String mobileLeft, String mobileRight) {
		
		String lrMobileSql = mobileLeft == null || mobileRight == null ? "" : String.format("DEVICE_NUMBER>='%s' and DEVICE_NUMBER<'%s'", mobileLeft, mobileRight);
		
		StrBuilder sqlBuilder = new StrBuilder();
		sqlBuilder.appendLn(super.buildTableSql(ZT_TABLE_NAME, null, "USERID"));
		sqlBuilder.append("select").appendLn(" USERID,");
		sqlBuilder.appendLn("if(CERT_AGE == '', '未知', CERT_AGE) CERT_AGE,")
				  .appendLn("if(PROVINCE == '', '未知', PROVINCE) PROVINCE,")
				  .appendLn("if(NET_TYPE == '', '未知', NET_TYPE) NET_TYPE,")
				  .appendLn("if(PRODUCT_CLASS == '', '未知', PRODUCT_CLASS) PRODUCT_CLASS,")
				  .appendLn("if(MONTH_FEE == '', '未知', MONTH_FEE) MONTH_FEE,")
				  .appendLn("if(MONTH_FLUX == '', '未知', MONTH_FLUX) MONTH_FLUX")
				  .appendLn("from(")
				  .appendLn("select DEVICE_NUMBER as USERID,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x) and x != '其他', groupArray(CERT_AGE_01)) as CERT_AGE,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x) and x != '其他', groupArray(PROV_DESC)) as PROVINCE,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x) and x != '无法区分', groupArray(SERVICE_TYPE)) as NET_TYPE,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x) and x != '其他主产品', groupArray(PRODUCT_CLASS_DESC)) as PRODUCT_CLASS,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x) and x != '其他', groupArray(mon_fee_01)) as MONTH_FEE,")
				  .appendLn("arrayFirst(x -> notEmpty(x) and isNotNull(x) and x != '其他', groupArray(total_flux_value_01)) as MONTH_FLUX")
				  .appendLn("from(")
				  .appendLn("with")
				  .appendLn("toInt32OrNull(SUBSTRING(MONTH_ID, 1, 4)) as month_value,")
				  .appendLn("toInt32OrNull(CERT_AGE) as age_value,")
				  .appendLn("toFloat64OrNull(MON_TOTAL_FEE) as mon_fee_value,")
				  .appendLn("toFloat64OrZero(TOTAL_FLUX) as total_flux_value")
				  .appendLn("select MONTH_ID,DEVICE_NUMBER, SERVICE_TYPE, PRODUCT_CLASS_DESC, PROV_DESC,")
				  .appendLn("IF(isNull(age_value), CERT_AGE, multiIf(month_value - age_value >= 2010, '10后', ")
				  .appendLn("month_value - age_value >= 2000, '00后', month_value - age_value >= 1990, '90后', month_value - age_value >= 1980, '80后',")
				  .appendLn("month_value - age_value >= 1970, '70后',month_value - age_value < 1970, '70前', '其他')) as CERT_AGE_01,")
				  .appendLn("if(isNULL(mon_fee_value), '其他', multiIf(mon_fee_value < 10, '10元以下', ")
				  .appendLn("mon_fee_value >= 10 and mon_fee_value<20, '10-20元', mon_fee_value >= 20 and mon_fee_value<60, '20-60元',")
				  .appendLn("mon_fee_value >= 60 and mon_fee_value<100, '60-100元',mon_fee_value >= 100 and mon_fee_value<150, '100-150元',")
				  .appendLn("mon_fee_value >= 150 and mon_fee_value<300, '150-300元',mon_fee_value >= 300, '300元以上','其他')) as mon_fee_01,")
				  .appendLn("if(isNULL(total_flux_value) or total_flux_value <= 0, '0',multiIf(total_flux_value > 0 and total_flux_value < 1024, '0-1G', ")
				  .appendLn("total_flux_value >= 1024 and total_flux_value<3072, '1-3G', total_flux_value >= 3072 and total_flux_value<5120, '3-5G',")
				  .appendLn("total_flux_value >= 5120 and total_flux_value<8192, '5-8G',total_flux_value >= 8192 and total_flux_value<10240, '8-10G',")
				  .appendLn("total_flux_value >= 10240 and total_flux_value<15360, '10-15G',total_flux_value >= 15360, '15G以上','其他')) as total_flux_value_01")
				  .appendLn("from XXH_ALL_MONTH where " + lrMobileSql + " order by MONTH_ID desc");
		sqlBuilder.appendLn(") group by DEVICE_NUMBER)");
		
//		log.info("\n" + sqlBuilder.toString());
		dynamicExecuteService.execute(DsEnum.clickhouse, sqlBuilder.toString(), String.format("[%s]用户部分标签的最新数据生成{%s-%s}", ZT_TABLE_NAME, mobileLeft, mobileRight));
	}
	
	@Override
	public boolean executeInBatchs() {
		return true;
	}
	
	@Override
	public long calcMobileRangeStep() {
		return _STEP;
	}
	
	@Override
	public void beforeInvoke(String month) {
		
	}

}
